/*
 *  Sample Code from SoundCodes~ Workshop
 *
 *  07. September 2016
 *  Copyright (c)2016 by Roald Christesen, Flensburg, Germany
 */

import javax.sound.sampled.*;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import ddf.minim.*;
import ddf.minim.ugens.*;


int sampleRate = 44100;              // *** Sample Rate, sinnvolle Werte 1000 - 96000
int sampleCount = sampleRate * 4;    // 4 Sekunden
int channelCount = 2;
int bitsPerSample = 24;
int bytesPerSample = 24 / 8;
static final int LEFT = 0;
static final int RIGHT = 1;


Minim minim;
AudioOutput output;
MultiChannelBuffer buffer;
Sampler sampler;





void pling() {

  // Parameter

  double freq1 = 220.0;        // *** Frequenz des ersten Oszillators, sinnvoll sind Werte zwischen 20.0 und 20000.0
  double freq2 = 221.0;        // *** Frequenz des zweiten Oszillators
  double freq3 = 439.0;        // *** Frequenz des dritten Oszillators
  
  double time1 = 4.0;          // *** Abklingzeit des ersten Oszillators in Sekunden, sinnvoll sind Werte zwischen 0.001 und 4.0
  double time2 = 2.0;          // *** Abklingzeit des zweiten Oszillators
  double time3 = 1.0;          // *** Abklingzeit des dritten Oszillators
  
  int bitDepth = 24;           // *** Die Bittiefe des fertigen Sounds, sinnvoll sind Werte von 1 - 24


  
  // Initialisierung
  
  minim  = new Minim(this);
  output = minim.getLineOut();

  buffer = new MultiChannelBuffer(sampleCount, channelCount);
  sampler = new Sampler(buffer, sampleRate, 1);
  sampler.patch(output);


  // Ab hier findet die Berechnung des Sounds statt ...
  
  double f1 = envelopeMultiplier(1, 0, sampleRate, time1);
  double f2 = envelopeMultiplier(1, 0, sampleRate, time2);
  double f3 = envelopeMultiplier(1, 0, sampleRate, time3);
  
  double a1 = 1.0;
  double a2 = 1.0;
  double a3 = 1.0;

  
  for (int sampleIndex = 0; sampleIndex < sampleCount; sampleIndex++) {

    float signal = 0.0;
    signal += (float)(sinus(freq1, sampleIndex) * a1);
    signal += (float)(sinus(freq2, sampleIndex) * a2);
    signal += (float)(sinus(freq3, sampleIndex) * a3);

    buffer.setSample(LEFT, sampleIndex, signal);
    buffer.setSample(RIGHT, sampleIndex, signal);
    
    a1 *= f1;
    a2 *= f2;
    a3 *= f3;
  }
   
  
  normalize();
  bitDepth(bitDepth);
}


void setup() {

  size(800, 400);
  
  pling();
}



void draw() {

  background(0);
  stroke(255);
  
  int waveHeight = (int)(height / 4 * 0.9);
  int c1 = height / 4;
  int c2 = height / 4 * 3;
  
  for (int i = 0; i < output.bufferSize() - 1; i++) {
    float x1 = map(i, 0, output.bufferSize(), 0, width);
    float x2 = map(i + 1, 0, output.bufferSize(), 0, width);
    line(x1, c1 - output.left.get(i) * waveHeight, x2, c1 - output.left.get(i + 1) * waveHeight);
    line(x1, c2 - output.right.get(i) * waveHeight, x2, c2 - output.right.get(i + 1) * waveHeight);
  }
}



void keyPressed() {

  try {
    
    if (sampler != null) {

      switch (key) {
        case ' ':
          sampler.trigger();
          break;

        case 's':
          saveToFile();
          break;
      }
    }
  }
  catch (IOException e) {
    System.out.println(e.getMessage());
  }
}



void saveToFile() throws IOException {
  
  File out = new File("pling.wav");
  boolean bigEndian = true;
  boolean signed = true;

  int size = sampleCount * bytesPerSample * channelCount;
  byte[] byteBuffer = new byte[size];

  int byteIndex = 0;
  for (int i = 0; i < sampleCount - 1; i++) {

    int value = (int)(buffer.getSample(LEFT, i) * 8388607);
    byteBuffer[byteIndex] = (byte)((value >> 16) & 0xFF);
    byteIndex++;
    byteBuffer[byteIndex] = (byte)((value >> 8) & 0xFF);
    byteIndex++;
    byteBuffer[byteIndex] = (byte)(value & 0xFF);
    byteIndex++;
    
    value = (int)(buffer.getSample(RIGHT, i) * 8388607);    
    byteBuffer[byteIndex] = (byte)((value >> 16) & 0xFF);
    byteIndex++;
    byteBuffer[byteIndex] = (byte)((value >> 8) & 0xFF);
    byteIndex++;
    byteBuffer[byteIndex] = (byte)(value & 0xFF);
    byteIndex++;
  }

  AudioFormat format = new AudioFormat((float)sampleRate, bitsPerSample, channelCount, signed, bigEndian);
  ByteArrayInputStream byteStream = new ByteArrayInputStream(byteBuffer);
  AudioInputStream audioInputStream;
  audioInputStream = new AudioInputStream(byteStream, format, sampleCount);
  AudioSystem.write(audioInputStream, AudioFileFormat.Type.WAVE, out);
  audioInputStream.close();
}


double randomSample() { return (Math.random() - 0.5) * 2; }
double sinus(double freq, int sampleIndex) { return Math.sin(Math.PI * 2 * freq / sampleRate * sampleIndex); }


void normalize() {
  
  float max = 0.0;
  for (int i = 0; i < sampleCount; i++) {
    for (int c = 0; c < channelCount; c++) {
      float value = Math.abs(buffer.getSample(c, i));
      if (value > max)
        max = value;
    }
  }

  if (max > 1.0) {
    float scale = 1.0 / max;
    
    for (int i = 0; i < sampleCount; i++) {
      for (int c = 0; c < channelCount; c++) {
        buffer.setSample(c, i, buffer.getSample(c, i) * scale);
      }
    }
  }
}



void bitDepth(int bit) {
  
  double f = Math.pow(2, bit - 1);
  
  for (int i = 0; i < sampleCount; i++) {
    double v = (double)((int)(buffer.getSample(LEFT, i) * f)) / f;
    buffer.setSample(LEFT, i, (float)v);
    v = (double)((int)(buffer.getSample(RIGHT, i) * f)) / f;
    buffer.setSample(RIGHT, i, (float)v);
  }
}


double bitDepth(double value, int bit) {
 
  double f = Math.pow(2, bit - 1);
  return (double)((int)(value * f)) / f;
}


double envelopeMultiplier(double startLevel, double endLevel, int sampleRate, double durationSeconds) {

  return envelopeMultiplier(startLevel, endLevel, (int)(durationSeconds * sampleRate));
}



double envelopeMultiplier(double startLevel, double endLevel, int sampleCount) {

  if (sampleCount < 1)
    return 1.0;
  
  if (startLevel < 0.0001)
    startLevel = 0.0001;

  if (endLevel < 0.0001)
    endLevel = 0.0001;
  
  return 1.0 + (Math.log(endLevel) - Math.log(startLevel)) / sampleCount;
}